@@ -36,9 +36,14 @@
 #include "task_states.h"
 #include "job_metadata.h"
 
-#include "poll.h"
+
+#ifdef HAVE_POLL_H
+#include <poll.h>
+#elif defined(HAVE_SYS_POLL_H)
+#include <sys/poll.h>
+#endif
+
 #include "sys/time.h"
-#include "sys/poll.h"
 #include "time.h"
 
 #include "access/genam.h"
@@ -133,7 +138,7 @@ static void ManageCronTasks(List *taskList, TimestampTz currentTime);
 static void ManageCronTask(CronTask *task, TimestampTz currentTime);
 static void ExecuteSqlString(const char *sql);
 static void GetTaskFeedback(PGresult *result, CronTask *task);
-static void GetBgwTaskFeedback(shm_mq_handle *responseq, CronTask *task);
+static void GetBgwTaskFeedback(shm_mq_handle *responseq, CronTask *task, bool running);
 
 static bool jobCanceled(CronTask *task);
 static bool jobStartupTimeout(CronTask *task, TimestampTz currentTime);
@@ -144,6 +149,7 @@ static void bgw_generate_returned_message(StringInfoData *display_msg, ErrorData
 char *CronTableDatabaseName = "postgres";
 static bool CronLogStatement = true;
 static bool CronLogRun = true;
+static bool CronReloadConfig = false;
 
 /* flags set by signal handlers */
 static volatile sig_atomic_t got_sigterm = false;
@@ -154,8 +160,27 @@ static const int MaxWait = 1000; /* maximum time in ms that poll() can block */
 static bool RebootJobsScheduled = false;
 static int RunningTaskCount = 0;
 static int MaxRunningTasks = 0;
+static int CronLogMinMessages = WARNING;
 static bool UseBackgroundWorkers = false;
 
+static const struct config_enum_entry cron_message_level_options[] = {
+	{"debug5", DEBUG5, false},
+	{"debug4", DEBUG4, false},
+	{"debug3", DEBUG3, false},
+	{"debug2", DEBUG2, false},
+	{"debug1", DEBUG1, false},
+	{"debug", DEBUG2, true},
+	{"info", INFO, false},
+	{"notice", NOTICE, false},
+	{"warning", WARNING, false},
+	{"error", ERROR, false},
+	{"log", LOG, false},
+	{"fatal", FATAL, false},
+	{"panic", PANIC, false},
+	{NULL, 0, false}
+};
+
+static const char *cron_error_severity(int elevel);
 
 /*
  * _PG_init gets called when the extension is loaded.
@@ -207,6 +232,16 @@ _PG_init(void)
 		GUC_SUPERUSER_ONLY,
 		NULL, NULL, NULL);
 
+	DefineCustomBoolVariable(
+		"cron.enable_superuser_jobs",
+		gettext_noop("Allow jobs to be scheduled as superuser"),
+		NULL,
+		&EnableSuperuserJobs,
+		true,
+		PGC_USERSET,
+		GUC_SUPERUSER_ONLY,
+		NULL, NULL, NULL);
+
 	DefineCustomStringVariable(
 		"cron.host",
 		gettext_noop("Hostname to connect to postgres."),
@@ -252,6 +287,17 @@ _PG_init(void)
 			GUC_SUPERUSER_ONLY,
 			NULL, NULL, NULL);
 
+	DefineCustomEnumVariable(
+		"cron.log_min_messages",
+		gettext_noop("log_min_messages for the launcher bgworker."),
+		NULL,
+		&CronLogMinMessages,
+		WARNING,
+		cron_message_level_options,
+		PGC_SIGHUP,
+		GUC_SUPERUSER_ONLY,
+		NULL, NULL, NULL);
+
 	/* set up common data for all our workers */
 	worker.bgw_flags = BGWORKER_SHMEM_ACCESS | BGWORKER_BACKEND_DATABASE_CONNECTION;
 	worker.bgw_start_time = BgWorkerStart_RecoveryFinished;
@@ -297,6 +343,7 @@ static void
 pg_cron_sighup(SIGNAL_ARGS)
 {
 	CronJobCacheValid = false;
+	CronReloadConfig = true;
 
 	if (MyProc != NULL)
 	{
@@ -361,6 +408,60 @@ interpret_error:
         return "";
 }
 
+/*
+ * cron_error_severity --- get string representing elevel
+ */
+static const char *
+cron_error_severity(int elevel)
+{
+	const char *elevel_char;
+
+	switch (elevel)
+	{
+		case DEBUG1:
+			elevel_char = "DEBUG1";
+			break;
+		case DEBUG2:
+			elevel_char = "DEBUG2";
+			break;
+		case DEBUG3:
+			elevel_char = "DEBUG3";
+			break;
+		case DEBUG4:
+			elevel_char = "DEBUG4";
+			break;
+		case DEBUG5:
+			elevel_char = "DEBUG5";
+			break;
+		case LOG:
+			elevel_char = "LOG";
+			break;
+		case INFO:
+			elevel_char = "INFO";
+			break;
+		case NOTICE:
+			elevel_char = "NOTICE";
+			break;
+		case WARNING:
+			elevel_char = "WARNING";
+			break;
+		case ERROR:
+			elevel_char = "ERROR";
+			break;
+		case FATAL:
+			elevel_char = "FATAL";
+			break;
+		case PANIC:
+			elevel_char = "PANIC";
+			break;
+		default:
+			elevel_char = "???";
+			break;
+	}
+
+	return elevel_char;
+}
+
 /*
  * bgw_generate_returned_message -
  *      generates the message to be inserted into the job_run_details table
@@ -516,6 +617,10 @@ PgCronLauncherMain(Datum arg)
 
 	ereport(LOG, (errmsg("pg_cron scheduler started")));
 
+	/* set the desired log_min_messages */
+	SetConfigOption("log_min_messages", cron_error_severity(CronLogMinMessages),
+										PGC_POSTMASTER, PGC_S_OVERRIDE);
+
 	MemoryContextSwitchTo(CronLoopContext);
 
 	while (!got_sigterm)
@@ -530,6 +635,15 @@ PgCronLauncherMain(Datum arg)
 			RefreshTaskHash();
 		}
 
+		if (CronReloadConfig)
+		{
+			/* set the desired log_min_messages */
+			ProcessConfigFile(PGC_SIGHUP);
+			SetConfigOption("log_min_messages", cron_error_severity(CronLogMinMessages),
+												PGC_POSTMASTER, PGC_S_OVERRIDE);
+			CronReloadConfig = false;
+		}
+
 		taskList = CurrentTaskList();
 		currentTime = GetCurrentTimestamp();
 
@@ -913,7 +1027,7 @@ PollForTasks(List *taskList)
 	int activeTaskCount = 0;
 	ListCell *taskCell = NULL;
 
-	polledTasks = (CronTask **) palloc0(taskCount * sizeof(CronTask));
+	polledTasks = (CronTask **) palloc0(taskCount * sizeof(CronTask *));
 	pollFDs = (struct pollfd *) palloc0(taskCount * sizeof(struct pollfd));
 
 	currentTime = GetCurrentTimestamp();
@@ -1560,10 +1674,6 @@ ManageCronTask(CronTask *task, TimestampTz currentTime)
 				break;
 			}
 
-			/* still waiting for job to complete */
-			if (GetBackgroundWorkerPid(&task->handle, &pid) != BGWH_STOPPED)
-				break;
-
 			toc = shm_toc_attach(PG_CRON_MAGIC, dsm_segment_address(task->seg));
 			#if PG_VERSION_NUM < 100000
 				mq = shm_toc_lookup(toc, PG_CRON_KEY_QUEUE);
@@ -1571,7 +1681,16 @@ ManageCronTask(CronTask *task, TimestampTz currentTime)
 				mq = shm_toc_lookup(toc, PG_CRON_KEY_QUEUE, false);
 			#endif
 			responseq = shm_mq_attach(mq, task->seg, NULL);
-			GetBgwTaskFeedback(responseq, task);
+
+			/* still waiting for job to complete */
+			if (GetBackgroundWorkerPid(&task->handle, &pid) != BGWH_STOPPED)
+			{
+				GetBgwTaskFeedback(responseq, task, true);
+				shm_mq_detach(responseq);
+				break;
+			}
+
+			GetBgwTaskFeedback(responseq, task, false);
 
 			task->state = CRON_TASK_DONE;
 			dsm_detach(task->seg);
@@ -1736,7 +1855,7 @@ GetTaskFeedback(PGresult *result, CronTask *task)
 }
 
 static void
-GetBgwTaskFeedback(shm_mq_handle *responseq, CronTask *task)
+GetBgwTaskFeedback(shm_mq_handle *responseq, CronTask *task, bool running)
 {
 
 	TimestampTz end_time;
@@ -1783,9 +1902,10 @@ GetBgwTaskFeedback(shm_mq_handle *responseq, CronTask *task)
 
 						if (edata.elevel >= ERROR)
 							UpdateJobRunDetail(task->runId, NULL, GetCronStatus(CRON_STATUS_FAILED), display_msg.data, NULL, &end_time);
+						else if (running)
+							UpdateJobRunDetail(task->runId, NULL, NULL, display_msg.data, NULL, NULL);
 						else
 							UpdateJobRunDetail(task->runId, NULL, GetCronStatus(CRON_STATUS_SUCCEEDED), display_msg.data, NULL, &end_time);
-
 					}
 
 					ereport(LOG, (errmsg("cron job " INT64_FORMAT ": %s",
